/*
 * Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *	Implementation of last write register
 *
 */


#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <complib/cl_last_write_register.h>

/******************************************************************************
 IMPLEMENTATION OF LAST WRITE REGISTER
******************************************************************************/

void cl_lwr_construct(IN cl_lwr_t * const p_lwr)
{
	CL_ASSERT(p_lwr);

	memset(p_lwr, 0, sizeof(*p_lwr));
	cl_spinlock_construct(&p_lwr->lock);
	p_lwr->state = CL_UNINITIALIZED;
}

cl_status_t cl_lwr_init(IN cl_lwr_t * const p_lwr,
			IN cl_pfn_lwr_dispose_t dispose,
			IN void * dispose_context)
{
	cl_status_t status;

	CL_ASSERT(p_lwr);
	CL_ASSERT(dispose);
	
	if (!cl_is_state_valid(p_lwr->state))
		cl_lwr_construct(p_lwr);

	status = cl_spinlock_init(&p_lwr->lock);
	if (status == CL_SUCCESS) {
		p_lwr->dispose = dispose;
		p_lwr->dispose_ctx = dispose_context;
		p_lwr->has_value = FALSE;
		p_lwr->state = CL_INITIALIZED;
	}

	return status;
}

cl_status_t cl_lwr_get(IN cl_lwr_t * const p_lwr, OUT void ** p_value)
{
	cl_status_t status;

	CL_ASSERT(p_lwr);
	CL_ASSERT(p_lwr->state == CL_INITIALIZED);
	CL_ASSERT(p_value);

	CL_SPINLOCK_ACQUIRE(&p_lwr->lock);
	/* State could change if someone called destroy() concurrently */
	if (p_lwr->state == CL_INITIALIZED) {
		if (p_lwr->has_value) {
			*p_value = p_lwr->last_value;
			status = CL_SUCCESS;
		} else {
			status = CL_NOT_FOUND;
		}
	} else {
		status = CL_INVALID_STATE;
	}
	CL_SPINLOCK_RELEASE(&p_lwr->lock);

	return status;
}

static cl_status_t __cl_lwr_pop_unsafe(IN cl_lwr_t * const p_lwr, OUT void ** p_value)
{
	if (p_lwr->has_value) {
		*p_value = p_lwr->last_value;
		p_lwr->has_value = FALSE;
		return CL_SUCCESS;
	} 
	
	return CL_NOT_FOUND;
}

cl_status_t cl_lwr_pop(IN cl_lwr_t * const p_lwr, OUT void ** p_value)
{
	cl_status_t status;

	CL_ASSERT(p_lwr);
	CL_ASSERT(p_lwr->state == CL_INITIALIZED);
	CL_ASSERT(p_value);

	CL_SPINLOCK_ACQUIRE(&p_lwr->lock);
	/* State could change if someone called destroy() concurrently */
	if (p_lwr->state == CL_INITIALIZED)
		status = __cl_lwr_pop_unsafe(p_lwr, p_value);
	else
		status = CL_INVALID_STATE;
	CL_SPINLOCK_RELEASE(&p_lwr->lock);

	return status;
}

void cl_lwr_put(IN cl_lwr_t * const p_lwr, IN void * value)
{
	void * previous_value;
	boolean_t had_value = FALSE;

	CL_ASSERT(p_lwr);
	CL_ASSERT(p_lwr->state == CL_INITIALIZED);

	CL_SPINLOCK_ACQUIRE(&p_lwr->lock);
	
	/* State could change if someone called destroy() concurrently */
	if (p_lwr->state == CL_INITIALIZED) {
		had_value = p_lwr->has_value;
		previous_value = p_lwr->last_value;
		p_lwr->last_value = value;
		p_lwr->has_value = TRUE;
	}
	
	CL_SPINLOCK_RELEASE(&p_lwr->lock);

	if (had_value)
		p_lwr->dispose(previous_value, p_lwr->dispose_ctx);
}

void cl_lwr_dispose_item(IN cl_lwr_t * const p_lwr, IN void * p_item)
{
	CL_ASSERT(p_lwr);
	CL_ASSERT(p_lwr->state == CL_INITIALIZED);

	p_lwr->dispose(p_item, p_lwr->dispose_ctx);
}

void cl_lwr_destroy(IN cl_lwr_t * const p_lwr)
{
	void * value;

	CL_ASSERT(p_lwr);
	CL_ASSERT(cl_is_state_valid(p_lwr->state));

	if (p_lwr->state == CL_INITIALIZED) {
		boolean_t do_dispose = FALSE;
		CL_SPINLOCK_ACQUIRE(&p_lwr->lock);
		
		p_lwr->state = CL_DESTROYING;

		if (__cl_lwr_pop_unsafe(p_lwr, &value) == CL_SUCCESS)
			do_dispose = TRUE;

		CL_SPINLOCK_RELEASE(&p_lwr->lock);

		if (do_dispose)
			p_lwr->dispose(value, p_lwr->dispose_ctx);
	}

	if (p_lwr->state != CL_DESTROYED) {
		cl_spinlock_destroy(&p_lwr->lock);
		p_lwr->state = CL_DESTROYED;
	}
}
